use smallset::SmallSet;

use crate::QVertex;

const CACHE_SIZE: usize = 256;
const BATCH_THRESHOLD: usize = 96;

struct VTFaceMap {
    face_count: Vec<u16>,
    offset_map: Vec<u32>,
    face_list : Vec<u32>,
}

impl VTFaceMap {
    fn faces(self: &VTFaceMap, vertex: u32) -> &[u32] {
        let offset = self.offset_map[vertex as usize] as usize;
        let count = self.face_count[vertex as usize] as usize;
        return &self.face_list[offset..offset+count]; // Rust ranges are end-exclusive (WTF)
    }

    fn create(faces: &[[u32;3]],max_vt: u32) -> VTFaceMap {
        let mut face_count = vec![0 as u16;max_vt as usize]; // if you connect 2^16 faces to a single vertex...
        for f in faces {
            // note: this causes problems with degenerate faces
            // but we shouldn't have those to begin with
            face_count[f[0] as usize] += 1;
            face_count[f[1] as usize] += 1;
            face_count[f[2] as usize] += 1;
        }
        let mut bsum : u32 = 0;
        let offset_map: Vec<u32> = face_count.iter().map(|v| {let tmp = bsum; bsum+=*v as u32; tmp}).collect();
        let mut cursor_map = offset_map.clone();
        let mut face_list = vec![0 as u32;bsum as usize];
        for (i,f) in faces.iter().enumerate() {
            // Have to match earlier loop exactly to not create problems
            face_list[cursor_map[f[0] as usize] as usize] = i as u32; cursor_map[f[0] as usize] += 1;
            face_list[cursor_map[f[1] as usize] as usize] = i as u32; cursor_map[f[1] as usize] += 1;
            face_list[cursor_map[f[2] as usize] as usize] = i as u32; cursor_map[f[2] as usize] += 1;
        }
    
        return VTFaceMap {face_count: face_count, offset_map: offset_map, face_list: face_list};
    }
}

#[derive(Debug, Copy, Clone)]
enum VTCacheState {
    // TODO: rename these stupid things
    Empty, // Undefined data in cache slot
    Busy(u32), // Vertex will be in cache slot after next fence
    Idle(u32), // Vertex has already been loaded
    Used(u32), // Vertex has been loaded and is being used -> locked
}

impl VTCacheState {
    fn cached_vt(self) -> Option<u32> {
        match self {
            VTCacheState::Busy(vi) | VTCacheState::Idle(vi) | VTCacheState::Used(vi) => Some(vi),
            _ => None,
        }
    }

    fn is_locked(self) -> bool {
        match self {
            VTCacheState::Busy(_) | VTCacheState::Used(_) => true,
            _ => false,
        }
    }

    fn on_fence(&mut self) {
        if let VTCacheState::Busy(vi) | VTCacheState::Used(vi) = self {
            *self = VTCacheState::Idle(*vi);
        }
    }

    fn on_use(&mut self) {
        if let VTCacheState::Idle(vi) = self {
            *self = VTCacheState::Used(*vi);
        }
    }
}

// Internal helper struct because even with Rust brr brr subfunctions are not a thing sometimes I hate high level languages
struct GeoOptimizeData<'a> {
    faces: &'a [[u32;3]],
    verts: &'a [QVertex],
    face_map : VTFaceMap,
    face_live: Vec<bool>,
    vt_live_faces: Vec<u16>,
    cache_vt : [VTCacheState;CACHE_SIZE],

    
}

impl GeoOptimizeData<'_> {


    fn init<'a>(faces: &'a [[u32;3]],verts: &'a [QVertex]) -> GeoOptimizeData<'a> {
        let face_map = VTFaceMap::create(&faces, verts.len().try_into().unwrap());
        return GeoOptimizeData {
            faces: faces,
            verts: verts,
            vt_live_faces: face_map.face_count.clone(),
            face_map: face_map,
            cache_vt: [VTCacheState::Empty;CACHE_SIZE],
            face_live: vec![true;faces.len()]
        }
    }

    fn vertex_cost(&self, v: u32) -> i32 {
        let mut live_faces = 0i32;
        let mut load_verts = SmallSet::<[u32;8]>::new();
        for fi in self.face_map.faces(v) {
            if self.face_live[*fi as usize] {
                live_faces += 1;
                for vi in self.faces[*fi as usize] {
                    if !load_verts.contains(&vi) && !self.cache_vt.iter().any(|vc| {vc.cached_vt() == Some(vi)}) {
                        load_verts.insert(vi);
                    }
                }
            }
        }
        // cost metric
        2*(load_verts.len() as i32) - 1*live_faces
    }

    fn replace_slot(&self, cur_face: [u32;3]) -> u8 {
        // Try finding an unused slot first
        self.cache_vt.iter().position(|vc| {
            match vc {
                VTCacheState::Empty => true,
                VTCacheState::Busy(_) | VTCacheState::Used(_) => false,
                VTCacheState::Idle(vi) => self.vt_live_faces[*vi as usize] == 0
            }
        }).or_else(|| {
            // Find highest cost idle vertex not part of current face
            self.cache_vt.iter().enumerate().flat_map(|(i,vc)| {
                match vc {
                    VTCacheState::Idle(vi) => if cur_face.contains(vi) {None} else {Some((i,*vi))}, 
                    _ => None
                }
            }).max_by_key(|(_i,vi)| {self.vertex_cost(*vi)}).map(|(i,_vi)|{i})
        }).expect("No idle cache slots").try_into().unwrap()
    }
}

fn cb_add_verts(cb: &mut Vec<u8>, verts: &[(u8,QVertex)]) {
    println!("push {} verts",verts.len());
    if verts.len() == 0 {return}
    cb.push(2); // magic number
    cb.push((verts.len()-1).try_into().expect("More than 256 verts?"));
    for (i,v) in verts {
        // todo: other vertex formats, less vec capacity check bloat
        cb.push(*i);
        cb.extend_from_slice(&v.x.to_le_bytes());
        cb.extend_from_slice(&v.y.to_le_bytes());
        cb.extend_from_slice(&v.z.to_le_bytes());
        cb.extend_from_slice(&v.nx.to_le_bytes());
        cb.extend_from_slice(&v.ny.to_le_bytes());
        cb.extend_from_slice(&v.nz.to_le_bytes());
        cb.extend_from_slice(&v.shade.to_le_bytes());
    }
}

fn cb_add_tris(cb: &mut Vec<u8>, tris: &[[u8;3]]) {
    println!("push {} tris",tris.len());
    if tris.len() == 0 {return}
    cb.push(3); // magic number
    cb.push((tris.len()-1).try_into().expect("More than 256 tris?"));
    for f in tris {
        cb.extend_from_slice(f);
    }
}

pub fn geo_encode(faces: &[[u32;3]],verts: &[QVertex]) -> Vec<u8> {

    let mut command_buffer = Vec::<u8>::new();
    let mut tri_pending = Vec::<[u8;3]>::new();
    let mut vt_pending = Vec::<(u8,QVertex)>::new();
    let mut vt_total = 0;
    let mut tri_total = 0;

    let mut data = GeoOptimizeData::init(faces, verts);

    // Vertex cache optimization algorithm
    // based on Lin and Yu's 2006 paper
    // "An improved vertex caching scheme for 3D mesh rendering."
    // IEEE transactions on visualization and computer graphics. 12. 640-8. 10.1109/TVCG.2006.59. 

    loop {
        // pick a focus vertex from current cache based on cost metric
        if let Some(v_focus) = data.cache_vt.iter().flat_map(|vc| {vc.cached_vt()}).
        filter(|v|{data.vt_live_faces[*v as usize] > 0}).min_by_key(|v| {data.vertex_cost(*v)}).or_else(||{
            // fallback for empty cache (pick first vt with a live face attached)
            data.vt_live_faces.iter().position(|l|{*l > 0}).map(|i|{i as u32})
        }) {
            // try emitting all faces around v_focus
            for fi in data.face_map.faces(v_focus) {
                if !data.face_live[*fi as usize] {continue;}
                let face = faces[*fi as usize];
                let mut vtc_ids = [0u8;3];
                for i in 0..3 { // Have to do this as a loop because closures explode the borrow check
                    let vi = face[i];
                    if let Some(load_slot) = data.cache_vt.iter().position(|vc| vc.cached_vt() == Some(vi)) {
                        vtc_ids[i] = load_slot as u8;
                        data.cache_vt[load_slot as usize].on_use();
                    } else {
                        let load_slot = data.replace_slot(face);
                        debug_assert!(!data.cache_vt[load_slot as usize].is_locked());
                        vt_pending.push((load_slot,verts[vi as usize].clone()));
                        vt_total += 1;
                        data.cache_vt[load_slot as usize] = VTCacheState::Busy(vi);
                        vtc_ids[i] = load_slot;
                    }
                }
                debug_assert!(vtc_ids.iter().all(|vci|{data.cache_vt[*vci as usize].is_locked()}));
                tri_pending.push(vtc_ids);
                tri_total += 1;
                for vi in face {data.vt_live_faces[vi as usize] -= 1};
                //println!("{} {} {} oof",data.vt_live_faces[v_focus as usize],data.face_map.face_count[v_focus as usize],data.face_map.offset_map[v_focus as usize]);
                data.face_live[*fi as usize] = false;

                if vt_pending.len() >= BATCH_THRESHOLD || tri_pending.len() == 256 {
                    cb_add_verts(&mut command_buffer, &vt_pending);
                    data.cache_vt.iter_mut().for_each(VTCacheState::on_fence);
                    vt_pending.clear();
                    cb_add_tris(&mut command_buffer, &tri_pending);
                    tri_pending.clear();
                }
            }
            // TODO: maybe see if any triangles can be rendered for free?
        } else {
            cb_add_verts(&mut command_buffer, &vt_pending);
            cb_add_tris(&mut command_buffer, &tri_pending);
            command_buffer.push(0); // magic number
            println!("{} total verts out, from {} in. {} total tris out, from {} in",vt_total,verts.len(),tri_total,faces.len());
            return command_buffer;
        }
    }

}

